package RIO.extaction;

import static rescuecore2.standard.entities.StandardEntityURN.BLOCKADE;
import static rescuecore2.standard.entities.StandardEntityURN.CIVILIAN;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

import com.google.common.collect.Lists;

import adf.agent.action.Action;
import adf.agent.action.ambulance.ActionLoad;
import adf.agent.action.ambulance.ActionRescue;
import adf.agent.action.ambulance.ActionUnload;
import adf.agent.action.common.ActionMove;
import adf.agent.action.common.ActionRest;
import adf.agent.communication.MessageManager;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.extaction.ExtAction;
import adf.component.module.algorithm.PathPlanning;
import rescuecore2.config.NoSuchConfigOptionException;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.AmbulanceTeam;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.worldmodel.EntityID;

public class RioneTransport extends ExtAction {
    private PathPlanning pathPlanning;

    private int thresholdRest;
    private int kernelTime;

    private HashMap<Integer, Action> actionHistory;
    private HashMap<Integer, Pair<Integer, Integer>> locationHistory;
    private HashMap<Integer, EntityID> positionHistory;
    private EntityID target;

    public RioneTransport(AgentInfo agentInfo, WorldInfo worldInfo, ScenarioInfo scenarioInfo, ModuleManager moduleManager, DevelopData developData) {
        super(agentInfo, worldInfo, scenarioInfo, moduleManager, developData);
        this.target = null;
        this.thresholdRest = developData.getInteger("ActionTransport.rest", 100);
        
        this.actionHistory = new HashMap<>();
        this.locationHistory = new HashMap<>();
        this.positionHistory = new HashMap<>();

        switch  (scenarioInfo.getMode()) {
            case PRECOMPUTATION_PHASE:
                this.pathPlanning = moduleManager.getModule("ActionTransport.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
            case PRECOMPUTED:
                this.pathPlanning = moduleManager.getModule("ActionTransport.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
            case NON_PRECOMPUTE:
                this.pathPlanning = moduleManager.getModule("ActionTransport.PathPlanning", "adf.sample.module.algorithm.SamplePathPlanning");
                break;
        }
    }

    public ExtAction precompute(PrecomputeData precomputeData) {
        super.precompute(precomputeData);
        if(this.getCountPrecompute() >= 2) {
            return this;
        }
        this.pathPlanning.precompute(precomputeData);
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    public ExtAction resume(PrecomputeData precomputeData) {
        super.resume(precomputeData);
        if(this.getCountResume() >= 2) {
            return this;
        }
        this.pathPlanning.resume(precomputeData);
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    public ExtAction preparate() {
        super.preparate();
        if(this.getCountPreparate() >= 2) {
            return this;
        }
        this.pathPlanning.preparate();
        try {
            this.kernelTime = this.scenarioInfo.getKernelTimesteps();
        }catch (NoSuchConfigOptionException e) {
            this.kernelTime = -1;
        }
        return this;
    }

    public ExtAction updateInfo(MessageManager messageManager){
        super.updateInfo(messageManager);
        if(this.getCountUpdateInfo() >= 2) {
            return this;
        }
        this.pathPlanning.updateInfo(messageManager);
        return this;
    }

    @Override
    public ExtAction setTarget(EntityID target) {
        this.target = null; //前回のtarget
        if(target != null) { //今のtarget
            StandardEntity entity = this.worldInfo.getEntity(target);
            if(entity instanceof Human || entity instanceof Area) {
                this.target = target;
                return this;
            }
        }
        return this;
    }
    
    

    @Override
    public ExtAction calc() {
        this.result = null;
        AmbulanceTeam agent = (AmbulanceTeam)this.agentInfo.me();
        StandardEntity position = this.worldInfo.getEntity(agentInfo.getPosition());        
        Human transportHuman = this.agentInfo.someoneOnBoard();
        if(transportHuman != null) {
            this.result = this.calcUnload(agent, transportHuman, this.target);
            if(this.result != null) {
            	this.setHistory(agent, this.result);
                return this;
            }
        }
        
        //ATのいる場所が燃えてる建物なら隣のエリアに移動する
        if(position instanceof Building && ((Building)position).isOnFire() &&
        		(((Building)position).getFierynessEnum() == Fieryness.INFERNO || ((Building)position).getFierynessEnum() == Fieryness.BURNING)){
        	this.pathPlanning.setFrom(position.getID());
        	this.pathPlanning.setDestination(((Building) position).getNeighbours());        	
        	List<EntityID> path = this.pathPlanning.calc().getResult();
        	if (path != null && path.size() > 0) {
                this.result = new ActionMove(path);  //Buildingからでる（隣のroad（getNeighbor））)
        	}
        	if(this.result != null){
        		this.setHistory(agent, this.result);
        		return this;
        	}
        }
        //this.resultが指定されない時
        if(this.needRest(agent)) {
            EntityID areaID = this.convertArea(this.target);  //(convert変換する)targetのposition（areaID）を返す
            ArrayList<EntityID> targets = new ArrayList<>();
            if(areaID != null) {
                targets.add(areaID);
            }
            this.result = this.calcRefugeAction(agent, targets, false);
            if(this.result != null) {
            	this.setHistory(agent, this.result);
                return this;
            }
        }
        
        if(this.target != null) {
            this.result = this.calcRescue(agent, this.target);
            if(this.result != null) {
            	this.setHistory(agent, this.result);
                return this;
            }
        }
        return this;
    }
    
    //動けない場合の処理
    private Action calcCannotMove(AmbulanceTeam ambulanceTeam, StandardEntity position) {
    	int time = this.agentInfo.getTime();
    	if(time > 5){
    		if(this.actionHistory.get(time-1) instanceof ActionMove){
    			if(this.equalsPoint(this.worldInfo.getLocation(ambulanceTeam), this.locationHistory.get(time-1))){
					Area moveArea = (Area)this.worldInfo.getEntity(((Area)position).getNeighbours().get(0));
					return new ActionMove(Lists.newArrayList(moveArea.getID()));
    			}
    		}
    	}
		return null;
	}
    
    //p1とp2が同じ点か調べる(100mmまでの誤差は許容しているので正確にはほぼ同じかどうか)
    private boolean equalsPoint(Pair<Integer, Integer> p1, Pair<Integer, Integer> p2) {
		return Math.hypot(p1.first()-p2.first(), p1.second()-p2.second()) < 300;
	}
    
	private List<EntityID> calcRefugeMoveAction(AmbulanceTeam ambulanceTeam){
		HashMap<List<EntityID>,Integer> costMap = new HashMap<>();
		List<StandardEntity> refuges = Lists.newArrayList(this.worldInfo.getEntitiesOfType(REFUGE));
		refuges.sort(new DistanceSorter(this.worldInfo, ambulanceTeam));
		int limit = 0;
		for(StandardEntity refuge : refuges) {
			int cost = 0;
			List<EntityID> path = this.pathPlanning.getResult(ambulanceTeam.getPosition(), refuge.getID());
			for(EntityID pathElementId : path){
				StandardEntity pathElement = this.worldInfo.getEntity(pathElementId);
				if(pathElement instanceof Road){
					Road road = (Road)pathElement;
					if(road.isBlockadesDefined() && road.getBlockades().size() > 0){
						for(Blockade blockade : this.worldInfo.getBlockades(road)){
							cost += blockade.getRepairCost();
						}
					}
				}
			}
			costMap.put(path, cost);
			limit++;
			if(limit > 2){
				break;
			}
		}
		
		int minCost = Integer.MAX_VALUE;
		List<EntityID> bestPath = null;
		for(Entry<List<EntityID>, Integer> entry : costMap.entrySet()){
			if(entry.getValue() < minCost){
				minCost = entry.getValue();
				bestPath = entry.getKey();
			}
		}
		if(bestPath != null){
			return bestPath;
		}		
		return null;
	}
    
    private Action calcRescue(AmbulanceTeam agent, EntityID targetID) {
        StandardEntity targetEntity = this.worldInfo.getEntity(targetID);
        if(targetEntity == null) {
            return null;
        }
        EntityID agentPosition = agent.getPosition();
        if(targetEntity instanceof Human) {
            Human human = (Human) targetEntity;
            if (!human.isPositionDefined()) {
                return null;
            }
            if (human.isHPDefined() && human.getHP() == 0) {
                return null;
            }
            EntityID targetPosition = human.getPosition();
            if (agentPosition.getValue() == targetPosition.getValue()) {
                if (human.isBuriednessDefined() && human.getBuriedness() > 0) {
                    return new ActionRescue(human);
                } else if (human.getStandardURN() == CIVILIAN) {
                    return new ActionLoad(human.getID());
                }
            } else {
                List<EntityID> path = this.pathPlanning.getResult(agentPosition, targetPosition);
                if (path != null && path.size() > 0) {
                    return new ActionMove(path);
                }
            }
            return null;
        }
        if(targetEntity.getStandardURN() == BLOCKADE) {
            Blockade blockade = (Blockade) targetEntity;
            if(blockade.isPositionDefined()) {
                targetEntity = this.worldInfo.getEntity(blockade.getPosition());
            }
        }
        if(targetEntity instanceof Area) {
            List<EntityID> path = this.pathPlanning.getResult(agentPosition, targetEntity.getID());
            if (path != null && path.size() > 0) {
                this.result = new ActionMove(path);
            }
        }
        return null;
    }
    
    private Action calcUnload(AmbulanceTeam agent, Human transportHuman, EntityID targetID) {
        if (transportHuman.isHPDefined() && transportHuman.getHP() == 0) {
            return new ActionUnload();
        }
        EntityID agentPosition = agent.getPosition();
        if(targetID == null || transportHuman.getID().getValue() == targetID.getValue()) {
            StandardEntity position = this.worldInfo.getEntity(agentPosition);
            if (position != null && position.getStandardURN() == REFUGE) {
                return new ActionUnload();
            } else {
            	List<EntityID> path = this.calcRefugeMoveAction(agent);
                //this.pathPlanning.setFrom(agentPosition);
                //this.pathPlanning.setDestination(this.worldInfo.getEntityIDsOfType(REFUGE));
                //List<EntityID> path = this.pathPlanning.calc().getResult();
                if (path != null && path.size() > 0) {
                    return new ActionMove(path);
                }
            }
        }
        StandardEntity targetEntity = this.worldInfo.getEntity(targetID);
        if(targetEntity != null && targetEntity.getStandardURN() == BLOCKADE) {
            Blockade blockade = (Blockade)targetEntity;
            if(blockade.isPositionDefined()) {
                targetEntity = this.worldInfo.getEntity(blockade.getPosition());
            }
        }
        if(targetEntity instanceof Area) {
            if (agentPosition.getValue() == targetID.getValue()) {
                return new ActionUnload();
            } else {
                this.pathPlanning.setFrom(agentPosition);
                this.pathPlanning.setDestination(targetID);
                List<EntityID> path = this.pathPlanning.calc().getResult();
                if (path != null && path.size() > 0) {
                    return new ActionMove(path);
                }
            }
        } else if(targetEntity instanceof Human) {
            Human human = (Human)targetEntity;
            if(human.isPositionDefined()) {
                return calcRefugeAction(agent, Lists.newArrayList(human.getPosition()), true);
            }
        	List<EntityID> path = this.calcRefugeMoveAction(agent);
            //this.pathPlanning.setFrom(agentPosition);
            //this.pathPlanning.setDestination(this.worldInfo.getEntityIDsOfType(REFUGE));
            //List<EntityID> path = this.pathPlanning.calc().getResult();
            if (path != null && path.size() > 0) {
                return new ActionMove(path);
            }
        }
        return null;
    }
    
    //action,positionを保存
    private void setHistory(AmbulanceTeam ambulanceTeam, Action action) {
		int time = this.agentInfo.getTime();
		this.actionHistory.put(time, action);
		this.locationHistory.put(time, this.worldInfo.getLocation(ambulanceTeam));
		this.positionHistory.put(time, ambulanceTeam.getPosition());
	}

    private boolean needRest(Human agent) {
        int hp = agent.getHP();
        int damage = agent.getDamage();
        if(hp == 0 || damage == 0) {
            return false;
        }
        int activeTime = (hp / damage) + ((hp % damage) != 0 ? 1 : 0);
        if(this.kernelTime == -1) {
            try {
                this.kernelTime = this.scenarioInfo.getKernelTimesteps();
            }catch (NoSuchConfigOptionException e) {
                this.kernelTime = -1;
            }
        }
        return damage >= this.thresholdRest || (activeTime + this.agentInfo.getTime()) < this.kernelTime;
    }

    private EntityID convertArea(EntityID targetID) {
        StandardEntity entity = this.worldInfo.getEntity(targetID);
        if(entity == null) {
            return null;
        }
        if(entity instanceof Human) {
            Human human = (Human) entity;
            if(human.isPositionDefined()) {
                EntityID position = human.getPosition();
                if(this.worldInfo.getEntity(position) instanceof Area) {
                    return position;
                }
            }
        }else if(entity instanceof Area) {
            return targetID;
        }else if(entity.getStandardURN() == BLOCKADE) {
            Blockade blockade = (Blockade)entity;
            if(blockade.isPositionDefined()) {
                return blockade.getPosition();
            }
        }
        return null;
    }

    private Action calcRefugeAction(AmbulanceTeam human, Collection<EntityID> targets, boolean isUnload) {
    	if(this.worldInfo.getEntity(human.getPosition()) instanceof Refuge){
    		return new ActionRest();
    	}
        List<EntityID> path = this.calcRefugeMoveAction(human);
        if(path != null && path.size() > 0){
        	return new ActionMove(path);
        }
        return null;
        /*EntityID position = human.getPosition();
        Collection<EntityID> refuges = this.worldInfo.getEntityIDsOfType(StandardEntityURN.REFUGE);
        int size = refuges.size();
        if(refuges.contains(position)) {
            return isUnload ? new ActionUnload() : new ActionRest();
        }
        List<EntityID> firstResult = null;
        while(refuges.size() > 0) {
            this.pathPlanning.setFrom(position);
            this.pathPlanning.setDestination(refuges);
            List<EntityID> path = this.pathPlanning.calc().getResult();
            if (path != null && path.size() > 0) {
                if (firstResult == null) {
                    firstResult = new ArrayList<>(path);
                    if(targets == null || targets.isEmpty()) {
                        break;
                    }
                }
                EntityID refugeID = path.get(path.size() - 1);
                this.pathPlanning.setFrom(refugeID);
                this.pathPlanning.setDestination(targets);
                List<EntityID> fromRefugeToTarget = this.pathPlanning.calc().getResult();
                if (fromRefugeToTarget != null && fromRefugeToTarget.size() > 0) {
                    return new ActionMove(path);
                }
                refuges.remove(refugeID);
                //remove failed
                if (size == refuges.size()) {
                    break;
                }
                size = refuges.size();
            } else {
                break;
            }
        }
        return firstResult != null ? new ActionMove(firstResult) : null;*/
    }
    
    private class DistanceSorter implements Comparator<StandardEntity> {
        private StandardEntity reference;
        private WorldInfo worldInfo;

        DistanceSorter(WorldInfo wi, StandardEntity reference) {
            this.reference = reference;
            this.worldInfo = wi;
        }

        public int compare(StandardEntity a, StandardEntity b) {
            int d1 = this.worldInfo.getDistance(this.reference, a);
            int d2 = this.worldInfo.getDistance(this.reference, b);
            return d1 - d2;
        }
    }
        
   /* 
    private Action cantMove() {
   	 PathPlanning pathPlanning = this.moduleManager.getModule("ActionTransport.PathPlanning");
        StandardEntity position = this.worldInfo.getEntity(agentInfo.getPosition());
        pathPlanning.setFrom(position.getID());
        
   	if(CommandHistory.get(agentInfo.getTime()-1) != null && CommandHistory.get(agentInfo.getTime()-2) != null){
   		if(CommandHistory.get(agentInfo.getTime()-1) instanceof ActionMove &&
       		CommandHistory.get(agentInfo.getTime()-2) instanceof ActionMove){
   			if(PositionHistory.get(agentInfo.getTime()-1) == PositionHistory.get(agentInfo.getTime()-2)){
   				System.out.println("0haru");//位置移動していなかったら
   				if (position instanceof Area) {	
   					pathPlanning.setDestination(((Area) position).getNeighbours());
   				    List<EntityID> path = pathPlanning.calc().getResult();
   				    System.out.println("1haru");
   				    if (path != null && path.size() > 0) {
  					        return new ActionMove(path);
  					    }
   				}
   			}
   		}else if(CommandHistory.get(agentInfo.getTime()-1) instanceof ActionRescue
   					&& position instanceof Building && ((Building)position).isOnFire()){
   			pathPlanning.setDestination(((Building) position).getNeighbours());
   			List<EntityID> path = pathPlanning.calc().getResult();
   			System.out.println("2haru");
   			if (path != null && path.size() > 0) {
   				return new ActionMove(path); //Buildingからでる（隣のroad（getNeighbor））)
   			}
   		}
   	}
   	return null;
	}*/ 
    
}
